#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <token.h>
#include <tokeniser.h>
#include <dynarray.h>

int parse_int(const char* s, int* out) {

    /* ----- handle optional sign ----- */
    long sign = 1;
    if (*s == '+' || *s == '-') {
        if (*s == '-') {
            sign = -1;
        }
        ++s;
    }
    if (*s == '\0') {
        return -1;
    } /* sign only? :3 */

    /* ----- binary special‑case ----- */
    if (s[0] == '0' && (s[1] == 'b' || s[1] == 'B')) {
        const char* p = s + 2;
        if (*p == '\0') {
            return -1;
        } /* “0b” but no digits */

        long val = 0;
        for (; *p; ++p) {
            if (*p == '0' || *p == '1') {
                val = (val << 1) | (*p - '0');
            } else {
                return -1;
            } /* stray char => error */
        }
        val *= sign;

        if (val < INT_MIN || val > INT_MAX) {
            return -1;
        }
        *out = (int)val;
        return 0;
    }

    /* ----- all other radices via strtol ----- */
    char* end;
    long val = strtol((sign == -1) ? (s - 1) : s, &end, 0); /* feed original string */
    if (*end != '\0') {
        return -1;
    } /* trailing junk */
    if (val < INT_MIN || val > INT_MAX) {
        return -1;
    }

    *out = (int)val;
    return 0;
}

token_t* tokenise_string(char* file_path, char* str) {
    const int str_len = strlen(str);
    loc_t loc = {0};
    loc.file = file_path;

    token_t* tokens = dynarray_create(token_t);

    for (int i = 0; i < str_len; i++) {
        switch (str[i]) {
        case (' '):
        case ('\t'):
        case ('\r'): {
            loc.col += 1;
            continue;
        }
        case ('\n'): {
            loc.col = 0;
            loc.line += 1;
            continue;
        }
        case ('c'):
        case ('"'): {
            if (str[i] == 'c') {
            }
            char* buf = malloc(1024 * 4);
            bool escaped = false;
            for (int y = 0;;) {
                if (str[i] == '"' && !escaped) {
                    break;
                }
                if (str[i] == '\\' && !escaped) {
                    escaped = true;
                    continue;
                }
                if (str[i] == '\\' && escaped) {
                    escaped = false;
                }
                buf[y++] = str[i++];
            }

            token_t tok = {
                .type = TT_PUSH_STR,
                .str_v = buf,
            };
            dynarray_push(tokens, tok);
        } break;
        default: {
            // TODO: manage memory better
            // INFO: If you make an ident longer than 4kb i will murder you
            char* buf = malloc(1024 * 4);
            memset(buf, 0, 1024 * 4);
            int buf_counter = 0;
            while (i < str_len) {
                if (str[i] == ' ' || str[i] == '\t' || str[i] == '\r' || str[i] == '\n') {
                    break;
                }
                buf[buf_counter++] = str[i];
                i += 1;
            }
            buf[buf_counter + 1] = '\0';
            printf("'%s'\n", buf);
            // PERF: I dont know if this makes it faster or slower, need 2 check
            buf = realloc(buf, strlen(buf) + 1);

            bool found = false;
            int num;
            if (parse_int(buf, &num) == 0) {
                token_t tok = {
                    .type = TT_PUSH_INT,
                    .int_v = num,
                };
                dynarray_push(tokens, tok);
                goto loop_end;
            }
            for (int i = 1; i < KW_COUNT__; i++) {
                if (strcmp(KW_LIST[i], buf) == 0) {
                    token_t tok = {
                        .type = TT_KW,
                        .kw_type = (kw_type_t)i,
                    };
                    dynarray_push(tokens, tok);
                    goto loop_end;
                }
            }

            for (int i = 1; i < OP_COUNT__; i++) {
                if (strcmp(OP_LIST[i], buf) == 0) {
                    token_t tok = {
                        .type = TT_OP,
                        .op_type = (op_type_t)i,
                    };
                    dynarray_push(tokens, tok);
                    goto loop_end;
                }
            }

            token_t tok = {
                .type = TT_IDENT,
                .str_v = strdup(buf),
            };
            dynarray_push(tokens, tok);
        loop_end:
            free(buf);
        }
        }
    }
    return tokens;
}
